import TurndownService from "turndown";
import interval from "interval-promise";
import { client, FORUMS_DOMAIN } from "../client";
import sha1 from "sha1";
import * as path from "path";
import * as fs from "fs";
import { HTML2BBCode } from "html2bbcode";
import { Dict } from "../util";
import { IAggregator, NewsPostItem } from "./aggregators/aggregator";
import { TextChannel, Message, Channel, ReactionCollector, MessageReaction, User, MessageEmbed } from "discord.js";
import { getRepository, IsNull, Not } from "typeorm";
import { KnownChannel } from "@shared/db/entity/KnownChannel";
import { AggroNewsItem } from "@shared/db/entity/AggroNewsItem";
import { v3beta1 } from "@google-cloud/translate";
import { logger } from "src/logging";
import { Plugin } from "src/model/plugin";
const { TranslationServiceClient } = v3beta1;

const UPDATE_INTERVAL = process.env.NODE_ENV == "dev" ? 60 : 5;
const MAX_PREVIEW_LENGTH = 300;
const AGGREGATOR_MANAGER_CHANNEL = "aggregatorManager";
const FORUMS_STAGING_ID = 54;
const FORUMS_NEWS_ID = 49;

@Plugin
export class NewsAggregator {
    tlClient = new TranslationServiceClient();
    aggregators: IAggregator[] = [];
    aggregateChannelID?: string;
    bbCodeParser = new HTML2BBCode();
    turndown = new TurndownService();
    reactionCollectors: Dict<ReactionCollector> = {};
    verifyMessageIdToPost: Dict<AggroNewsItem> = {};

    constructor() {
        this.turndown.addRule("image", {
            filter: "img",
            replacement: () => ""
        });
        this.turndown.addRule("link", {
            filter: (node: HTMLElement) => node.nodeName === "A" && node.getAttribute("href") != null,
            replacement: (content: string, node: Node) => (node as HTMLElement).getAttribute("htef") ?? ""
        });
    }

    checkFeeds = async (): Promise<void> => {
        logger.info(`Aggregating feeds on ${new Date().toISOString()}`);

        const aggregatorJobs = [];

        for (const aggregator of this.aggregators) {
            aggregatorJobs.push(aggregator.aggregate());
        }
        const aggregatedItems = await Promise.all(aggregatorJobs);

        for (const itemSet of aggregatedItems) {
            for (const item of itemSet) {
                const itemObj = {
                    ...item,
                    cacheMessageId: undefined,
                    postedMessageId: undefined
                } as NewsPostItem;

                itemObj.hash = sha1(itemObj.contents);

                await this.addNewsItem(itemObj);
            }
        }
    }

    clipText(text: string): string {
        if (text.length <= MAX_PREVIEW_LENGTH)
            return text;

        return `${text.substring(0, MAX_PREVIEW_LENGTH)}...`;
    }

    async addNewsItem(item: NewsPostItem): Promise<void> {
        if (!this.aggregateChannelID)
            return;

        const repo = getRepository(AggroNewsItem);

        const ch = client.bot.channels.resolve(this.aggregateChannelID);

        if (!(ch instanceof TextChannel))
            return;

        let isNew = true;
        let newsItem = await repo.findOne({
            where: { feedName: item.feedId, newsId: item.newsId }
        });

        if (newsItem) {
            if (process.env.IGNORE_CHANGED_NEWS === "TRUE") {
                newsItem.hash = item.hash ?? "";
                await repo.save(newsItem);
                return;
            }

            // No changes, skip
            if (newsItem.hash == item.hash)
                return;
            else if(newsItem.editMessageId)
                await this.deleteCacheMessage(newsItem.editMessageId);
            isNew = false;
        } else {
            newsItem = repo.create({
                newsId: item.newsId,
                feedName: item.feedId,
                hash: item.hash
            });
        }

        if (item.needsTranslation && process.env.GOOGLE_APP_ID)
            try {
                const request = {
                    parent: this.tlClient.locationPath(process.env.GOOGLE_APP_ID, "global"),
                    contents: [item.title, item.contents],
                    mimeType: "text/html",
                    sourceLanguageCode: "ja",
                    targetLanguageCode: "en"
                };

                const [res] = await this.tlClient.translateText(request);
                const [translatedTitle, translatedContents] = res.translations ?? [undefined, undefined];
                
                if (translatedTitle?.translatedText && translatedContents?.translatedText) {
                    item.title = translatedTitle.translatedText;
                    item.contents = translatedContents.translatedText;
                }
            } catch (err) {
                logger.error("Failed to translate news with Google", err);
            }

        item.contents = this.bbCodeParser.feed(item.contents).toString();

        if (!newsItem.forumsEditPostId) {
            const createResponse = await client.forum.createThread(FORUMS_STAGING_ID, item.title, item.contents);
            newsItem.forumsEditPostId = createResponse.thread.thread_id;
        } else if(newsItem.forumsNewsPostId){
            await client.forum.postReply(newsItem.forumsNewsPostId, item.contents);
        }


        const msg = await ch.send(new MessageEmbed({
            title: item.title,
            url: item.link,
            color: item.embedColor,
            timestamp: new Date(),
            description: `${(isNew ? "**[NEW]**" : "**[EDIT]**")}\n[**Edit on forums**](${FORUMS_DOMAIN}/index.php?threads/.${newsItem.forumsEditPostId}/)`,
            author: {
                name: item.author
            },
            footer: {
                text: "NoctBot News Aggregator"
            }
        })) as Message;

        newsItem.editMessageId = msg.id;

        await msg.react("✅");
        await msg.react("❌");

        const collector = msg.createReactionCollector(this.isVerifyReaction, { maxEmojis: 1 });
        collector.on("collect", this.collectorFor(collector));
        this.reactionCollectors[msg.id] = collector;
        this.verifyMessageIdToPost[msg.id] = newsItem;

        await repo.save(newsItem);
    }

    isVerifyReaction(reaction: MessageReaction, user: User): boolean {
        return (reaction.emoji.name == "✅" || reaction.emoji.name == "❌") && !user.bot && user.id != client.botUser.id;
    }

    collectorFor = (collector: ReactionCollector) =>
        async (reaction: MessageReaction): Promise<void> => {
            const repo = getRepository(AggroNewsItem);

            const m = reaction.message;
            collector.stop();
            delete this.reactionCollectors[m.id];
            const post = this.verifyMessageIdToPost[m.id];

            if(!post.forumsEditPostId) {
                throw new Error("No forum edit post found!");
            }

            if (reaction.emoji.name == "✅") {
                const res = await client.forum.getThread(post.forumsEditPostId);
                const forumPost = await client.forum.getPost(res.thread.first_post_id);

                if (!post.forumsNewsPostId) {
                    const newThread = await client.forum.createThread(FORUMS_NEWS_ID, res.thread.title, forumPost.message);
                    post.forumsNewsPostId = newThread.thread.thread_id;
                } else {
                    const curThread = await client.forum.editThread(post.forumsNewsPostId, {
                        title: res.thread.title
                    });

                    await client.forum.editPost(curThread.thread.first_post_id, {
                        message: forumPost.message
                    });
                }
            }

            await client.forum.deleteThread(post.forumsEditPostId);
            await repo.update({ newsId: post.newsId, feedName: post.feedName }, { 
                editMessageId: undefined, 
                forumsEditPostId: undefined, 
                forumsNewsPostId: post.forumsNewsPostId });
            await reaction.message.delete();
            delete this.verifyMessageIdToPost[m.id];
        };

    async deleteCacheMessage(messageId: string): Promise<void> {
        if(!this.aggregateChannelID)
            return;
        const ch = client.bot.channels.resolve(this.aggregateChannelID);
        if (!(ch instanceof TextChannel))
            return;

        const msg = await this.tryFetchMessage(ch, messageId);

        if (msg)
            await msg.delete();
    }

    async tryFetchMessage(channel: Channel, messageId: string): Promise<Message | null> {
        try {
            if (!(channel instanceof TextChannel))
                return null;
            return await channel.messages.fetch(messageId);
        } catch (error) {
            return null;
        }
    }

    initAggregators(): void {
        const aggregatorsPath = path.join(path.dirname(module.filename), "aggregators");
        const files = fs.readdirSync(aggregatorsPath);

        for (const file of files) {
            const ext = path.extname(file);
            const name = path.basename(file);

            if (name == "aggregator.js")
                continue;
            if (ext != ".js")
                continue;

            // eslint-disable-next-line @typescript-eslint/no-var-requires
            const obj = require(path.resolve(aggregatorsPath, file)).default as IAggregator;

            if (obj)
                this.aggregators.push(obj);

            if (obj.init)
                obj.init();
        }
    }

    async initPendingReactors(): Promise<void> {
        if(!this.aggregateChannelID)
            return;

        const verifyChannel = client.bot.channels.resolve(this.aggregateChannelID);
        if(!verifyChannel)
            throw new Error("Couldn't find verify channel!");

        const repo = getRepository(AggroNewsItem);

        const pendingVerifyMessages = await repo.find({
            where: { editMessageId: Not(IsNull()) }
        });

        for (const msg of pendingVerifyMessages) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            const m = await this.tryFetchMessage(verifyChannel, msg.editMessageId!);

            if (!m) {
                await repo.update({ feedName: msg.feedName, newsId: msg.newsId }, { editMessageId: undefined });
                continue;
            }

            const collector = m.createReactionCollector(this.isVerifyReaction, { maxEmojis: 1 });
            collector.on("collect", this.collectorFor(collector));
            this.reactionCollectors[m.id] = collector;
            this.verifyMessageIdToPost[m.id] = msg;
        }
    }

    async start(): Promise<void> {
        const repo = getRepository(KnownChannel);

        const ch = await repo.findOne({
            where: { channelType: AGGREGATOR_MANAGER_CHANNEL }
        });

        if (!ch)
            return;

        this.aggregateChannelID = ch.channelId;

        await this.initPendingReactors();
        this.initAggregators();
        interval(this.checkFeeds, UPDATE_INTERVAL * 60 * 1000);
    }
}